#define __STDC_CONSTANT_MACROS
#include <iostream>
#include <map>
#include <algorithm>
#include <iterator>
#include <cassert>
#include <cmath>
#include <stdint.h>
#include "arith.h"

void context::code_set(const std::vector<int> &codes)
{
    assert(freq.empty());
    int low = *min_element(codes.begin(), codes.end());
    int high = *max_element(codes.begin(), codes.end());
    offset = low;
    range = 1;
    presets = 0;
    while (range < high - low + 1) range *= 2;
    freq.resize(2 * range, 0);
    for (std::vector<int>::const_iterator i = codes.begin(); i != codes.end(); i++)
    {
        int v = *i - offset;
        int x = v + range;
        if (freq[x]) continue;
        while (x)
        {
            freq[x]++;
            x >>= 1;
        }
        presets++;
    }
}

void context::code_range(int low, int high)
{
    assert(freq.empty());
    assert(low <= high);
    offset = low;
    range = 1;
    while (range < high - low + 1) range *= 2;
    presets = high - low + 1;
    freq.resize(2 * range, 0);
    for (size_t i = range; i < range + presets; i++)
    {
        int x = i;
        while (x)
        {
            freq[x]++;
            x >>= 1;
        }
    }
}

std::size_t context::use() const
{
    assert(!freq.empty());
    return freq[1] - presets;
}

double context::entropy() const
{
    std::size_t total = use();
    double e = 0.0;

    for (int i = range; i < 2 * range; i++)
        if (freq[i] > 1)
        {
            double p = (freq[i] - 1) / (double) total;
            e += p * log(p) / log(0.5);
        }
    return e * total;
}

std::ostream &operator <<(std::ostream &o, const context &ctx)
{
    for (int i = ctx.range; i < 2 * ctx.range; i++)
        if (ctx.freq[i] > 1)
            o << i - ctx.range + ctx.offset << ": " << ctx.freq[i] - 1 << "\n";
    return o;
}

compressor::compressor()
: buffer(0), buffer_bits(0), total_bits(0.0), quarter_bits(0),
low(0), high(UINT64_C(0x100000000)), out(NULL)
{
}

void compressor::open(std::ostream &o)
{
    out = &o;
}

void compressor::close()
{
    if (low != high)
    {
        for (int i = 0; i < 32; i++)
        {
            output_bit((low & UINT64_C(0x80000000)) ? 1 : 0);
            low = (low << 1) & UINT64_C(0xffffffff);
        }
        while (quarter_bits || buffer_bits) output_bit(0);
        high = low;
    }
}

void compressor::output_raw_bit(int b)
{
    buffer <<= 1;
    buffer |= b;
    buffer_bits++;
    if (buffer_bits == 8)
    {
        if (out) out->write((char *) &buffer, 1);
        buffer_bits = 0;
        buffer = 0;
    }
}

void compressor::output_bit(int b)
{
    output_raw_bit(b);
    for (int i = 0; i < quarter_bits; i++)
        output_raw_bit(!b);
    quarter_bits = 0;
}

double compressor::encode_raw(int _bottom, int _top, int _total)
{
    uint64_t bottom = _bottom, top = _top, total = _total, r;

    double b = log((top - bottom) / (double) total) / log(0.5);
    r = high - low;
    high = low + top * r / total;
    low = low + bottom * r / total;
    while (high - low <= UINT64_C(0x40000000))
    {
        if (high <= UINT64_C(0x80000000))
        {
            output_bit(0);
            high <<= 1;
            low <<= 1;
        }
        else if (low >= UINT64_C(0x80000000))
        {
            output_bit(1);
            high -= UINT64_C(0x80000000); high <<= 1;
            low -= UINT64_C(0x80000000); low <<= 1;
        }
        else
        {
            quarter_bits++;
            low -= UINT64_C(0x40000000); low <<= 1;
            high -= UINT64_C(0x40000000); high <<= 1;
        }
    }

    total_bits += b;
    return b;
}

double compressor::encode(context &ctx, int value)
{
    int bottom = 0, top = 0;
    int id = value - ctx.offset;
    int inx = id + ctx.range;

    assert(id >= 0 && id < ctx.range);
    assert(ctx.freq[inx]);
    while (inx > 1)
    {
        if (inx & 1) bottom += ctx.freq[inx - 1];
        inx >>= 1;
    }
    top = bottom + ctx.freq[id + ctx.range];
    double b = encode_raw(bottom, top, ctx.freq[1]);
    ctx.total_bits += b;

    inx = id + ctx.range;
    while (inx)
    {
        ctx.freq[inx]++;
        inx >>= 1;
    }
    return b;
}

double compressor::encode_uniform(int value, int range)
{
    return encode_raw(value, value + 1, range);
}

compressor::~compressor()
{
    close();
}

void decompressor::input_bit()
{
    int bit;
    if (!buffer_bits)
    {
        input.read((char *) &buffer, 1);
        buffer_bits = 8;
    }
    bit = (buffer & 0x80) ? 1 : 0;
    buffer <<= 1;
    buffer_bits--;

    low <<= 1;
    high <<= 1;
    value = (value << 1) + bit;
}

void decompressor::input_bits()
{
    while (high - low <= UINT64_C(0x40000000))
    {
        if (high <= UINT64_C(0x80000000))
            input_bit();
        else if (low >= UINT64_C(0x80000000))
        {
            high -= UINT64_C(0x80000000);
            low -= UINT64_C(0x80000000);
            value -= UINT64_C(0x80000000);
            input_bit();
        }
        else
        {
            low -= UINT64_C(0x40000000);
            high -= UINT64_C(0x40000000);
            value -= UINT64_C(0x40000000);
            input_bit();
        }
    }
}

decompressor::decompressor(std::istream &in)
: input(in), buffer(0), buffer_bits(0), low(0), high(1), value(0)
{
    while (high != UINT64_C(0x100000000)) input_bit();
}

int decompressor::decode(context &ctx)
{
    assert(low <= value && value < high);
    assert(!ctx.freq.empty());

    uint64_t total = ctx.freq[1], test;
    uint64_t range = high - low;
    std::size_t sum = 0;
    int inx = 1;

    while (inx < ctx.range)
    {
        inx *= 2;
        test = low + (sum + ctx.freq[inx]) * range / total;
        if (test <= value)
        {
            sum += ctx.freq[inx];
            inx++;
        }
    }

    high = low + (sum + ctx.freq[inx]) * range / total;
    low = low + sum * range / total;
    input_bits();

    int id = inx - ctx.range;
    while (inx)
    {
        ctx.freq[inx]++;
        inx >>= 1;
    }
    return id + ctx.offset;
}

int decompressor::decode_uniform(int range)
{
    uint64_t r = high - low;
    uint64_t v;

    v = ((value - low) * range + range - 1) / r;
    high = low + (v + 1) * r / range;
    low = low + v * r / range;
    input_bits();
    return v;
}

decompressor::~decompressor()
{
}
