#include <valarray>
#include <vector>
#include <istream>
#include <ostream>
#include <cstddef>
#include <algorithm>
#include <cmath>
#include <stdint.h>
#include "arith.h"
#include "coders3d.h"

/****************
 * Utility code *
 ****************/
static inline double length(const std::valarray<int> &x)
{
    return sqrt((double) x[0] * (double) x[0]
                + (double) x[1] * (double) x[1]
                + (double) x[2] * (double) x[2]);
}

static inline int round_length(const std::valarray<int> &x)
{
    return (int) round(length(x));
}

/************************************************************
 * Axial coder: X, Y and Z are encoded in separate contexts *
 ************************************************************/

axial_coder::axial_coder(int max_x, int max_y, int max_z)
: max_x(max_x), max_y(max_y), max_z(max_z)
{
    axes[0].code_range(-max_x, max_x);
    axes[1].code_range(-max_y, max_y);
    axes[2].code_range(-max_z, max_z);
}

axial_coder::axial_coder(std::istream &in)
{
    uint32_t mx, my, mz;
    in.read((char *) &mx, sizeof(mx));
    in.read((char *) &my, sizeof(my));
    in.read((char *) &mz, sizeof(mz));
    max_x = mx; max_y = my; max_z = mz;
    axes[0].code_range(-max_x, max_x);
    axes[1].code_range(-max_y, max_y);
    axes[2].code_range(-max_z, max_z);
}

double axial_coder::encode(compressor &c, const std::valarray<int> &d)
{
    double b = 0.0;
    b += c.encode(axes[0], d[0]);
    b += c.encode(axes[1], d[1]);
    b += c.encode(axes[2], d[2]);
    return b;
}

std::valarray<int> axial_coder::decode(decompressor &d)
{
    std::valarray<int> ans(3);
    ans[0] = d.decode(axes[0]);
    ans[1] = d.decode(axes[1]);
    ans[2] = d.decode(axes[2]);
    return ans;
}

double axial_coder::bits() const
{
    return axes[0].bits() + axes[1].bits() + axes[2].bits();
}

double axial_coder::entropy() const
{
    return axes[0].entropy() + axes[1].entropy() + axes[2].entropy();
}

std::size_t axial_coder::use() const
{
    return axes[0].use();
}

void axial_coder::dump(std::ostream &out) const
{
    out << "X:\n" << axes[0] << "Y:\n" << axes[1] << "Z:\n" << axes[2];
}

void axial_coder::write_header(std::ostream &out) const
{
    uint32_t mx = max_x;
    uint32_t my = max_y;
    uint32_t mz = max_z;
    out.write((const char *) &mx, sizeof(mx));
    out.write((const char *) &my, sizeof(my));
    out.write((const char *) &mz, sizeof(mz));
}

/*******************************************************************
 * Single context coder: similar to axial_coder, but uses only one *
 * context for coordinates.                                        *
 *******************************************************************/

single_coder::single_coder(int max_x, int max_y, int max_z)
{
    max_xyz = std::max(max_x, std::max(max_y, max_z));
    freq.code_range(-max_xyz, max_xyz);
}

single_coder::single_coder(std::istream &in)
{
    uint32_t m;
    in.read((char *) &m, sizeof(m));
    max_xyz = m;
    freq.code_range(-m, m);
}

double single_coder::encode(compressor &c, const std::valarray<int> &d)
{
    double b = 0.0;
    b += c.encode(freq, d[0]);
    b += c.encode(freq, d[1]);
    b += c.encode(freq, d[2]);
    return b;
}

std::valarray<int> single_coder::decode(decompressor &d)
{
    std::valarray<int> ans(3);
    ans[0] = d.decode(freq);
    ans[1] = d.decode(freq);
    ans[2] = d.decode(freq);
    return ans;
}

double single_coder::bits() const
{
    return freq.bits();
}

double single_coder::entropy() const
{
    return freq.entropy();
}

std::size_t single_coder::use() const
{
    return freq.use() / 3;
}

void single_coder::dump(std::ostream &out) const
{
    out << "freq:\n" << freq;
}

void single_coder::write_header(std::ostream &out) const
{
    uint32_t m = max_xyz;
    out.write((const char *) &m, sizeof(m));
}

/***************************************************************
 * Radial coder: encodes (int) radius, then index with uniform *
 ***************************************************************/

radial_coder::radial_coder(int max_length)
:
max_length(max_length),
rfreq(),
sizes(),
index(),
total_bits(0.0), radial_entropy(0.0)
{
    initialise(true);
}

radial_coder::radial_coder(std::istream &in)
:
max_length(0),
rfreq(),
sizes(),
index(),
total_bits(0.0), radial_entropy(0.0)
{
    uint32_t ml;

    in.read((char *) &ml, sizeof(ml));
    max_length = ml;
    initialise(false);
}

void radial_coder::initialise(bool compressing)
{
    sizes.resize(max_length + 1, 0);
    if (compressing)
        index.resize(2 * max_length + 1, std::vector<std::vector<int> >(2 * max_length + 1, std::vector<int>(2 * max_length + 1, 0)));
    else
        rindex.resize(max_length + 1);
    rfreq.code_range(0, max_length);

    std::valarray<int> xyz(3);
    for (int z = -max_length; z <= max_length; z++)
        for (int y = -max_length; y <= max_length; y++)
            for (int x = -max_length; x <= max_length; x++)
            {
                xyz[0] = x;
                xyz[1] = y;
                xyz[2] = z;
                int l = round_length(xyz);
                if (l <= max_length)
                {
                    radial_address v;
                    v.x = x; v.y = y; v.z = z;
                    if (compressing)
                        index[z + max_length][y + max_length][x + max_length] = sizes[l];
                    else
                        rindex[l].push_back(v);
                    sizes[l]++;
                }
            }
}

double radial_coder::encode(compressor &c, const std::valarray<int> &d)
{
    int l = round_length(d);
    std::valarray<int> ind(3);
    double b = 0.0;

    ind[0] = d[0] + max_length;
    ind[1] = d[1] + max_length;
    ind[2] = d[2] + max_length;
    b += c.encode(rfreq, l);
    b += c.encode_uniform(index[ind[2]][ind[1]][ind[0]], sizes[l]);
    radial_entropy += log((double) sizes[l]) / log(2.0);
    total_bits += b;
    return b;
}

std::valarray<int> radial_coder::decode(decompressor &d)
{
    std::valarray<int> ans(3);
    int l = d.decode(rfreq);
    int inx = d.decode_uniform(sizes[l]);
    ans[0] = rindex[l][inx].x;
    ans[1] = rindex[l][inx].y;
    ans[2] = rindex[l][inx].z;
    return ans;
}

double radial_coder::bits() const
{
    return total_bits;
}

double radial_coder::entropy() const
{
    return radial_entropy + rfreq.entropy();
}

std::size_t radial_coder::use() const
{
    return rfreq.use();
}

void radial_coder::dump(std::ostream &out) const
{
    out << "Lengths:\n" << rfreq << "\n";
}

void radial_coder::write_header(std::ostream &out) const
{
    uint32_t ml = max_length;
    out.write((const char *) &ml, sizeof(ml));
}
