#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <queue>
#include <map>
#include <valarray>
#include <cstdlib>
#include <climits>
#include <cmath>
#include <cassert>
#include <unistd.h>
#include "arith.h"

enum choice
{
    ALWAYS,
    NEVER,
    MAYBE
};

enum universe
{
    CTX_PLAIN,
    CTX_FLIP,
    CTX_COUNT
};

using namespace std;

int slew = 0;
bool full_mst = false;
choice flipflop = MAYBE;
bool use_conditional = true;
bool reorient = true;
bool basis = false;
bool max_metric = false;
string dump_file = "";

struct edge
{
    int target;

    edge() {}
    explicit edge(int t) : target(t) {}
};

struct node
{
    valarray<double> rcoord;
    valarray<int> coord;
    vector<edge> edges;

    node() : rcoord(3), coord(3), edges() {}
};

struct flip_vector
{
    valarray<int> value;
    int ctx;

    flip_vector() : value(3), ctx(0) {}
};

class encoder
{
public:
    virtual void output(compressor &c, const valarray<int> &d) = 0;
    virtual ~encoder() {}
};

class axial_encoder : public encoder
{
private:
    context axes[3];

public:
    axial_encoder(int max_length);
    virtual void output(compressor &c, const valarray<int> &d);
    virtual ~axial_encoder() {}
};

class conditional_encoder : public encoder
{
private:
    context xfreq;
    map<int, context > xyfreq;
    map<int, context > xzfreq;
    bool lg;

public:
    conditional_encoder(int max_length, bool lg);
    virtual void output(compressor &c, const valarray<int> &d);
    virtual ~conditional_encoder() {}
};

vector<node> nodes;
vector<int> parent;
vector<int> order;
vector<valarray<double> > normal;
vector<valarray<int> > offsets[2];

template<class T>
static ostream &operator <<(ostream &o, const valarray<T> d)
{
    o << d[0] << " " << d[1] << " " << d[2];
    return o;
}

static void readin(const char *fname)
{
    string line;
    int nv, nf;

    ifstream f(fname);
    while (getline(f, line))
    {
        string token;
        istringstream s(line);

        s >> token;
        if (token == "element")
        {
            s >> token;
            if (token == "vertex")
                s >> nv;
            else if (token == "face")
                s >> nf;
            else
                abort();
        }
        else if (token == "end_header") break;
    }

    nodes.resize(nv);
    for (int i = 0; i < nv; i++)
    {
        getline(f, line);
        istringstream s(line);

        s >> nodes[i].rcoord[0]
            >> nodes[i].rcoord[1]
            >> nodes[i].rcoord[2];
    }
    for (int i = 0; i < nf; i++)
    {
        getline(f, line);
        istringstream s(line);
        int c;
        vector<int> id;

        s >> c;
        id.resize(c);
        for (int i = 0; i < c; i++)
            s >> id[i];
        for (int i = 0; i < c; i++)
        {
            int j = (i + 1) % c;
            nodes[id[i]].edges.push_back(edge(id[j]));
        }
    }
}

static void quantise(int bits)
{
    valarray<double> mn = nodes[0].rcoord;
    valarray<double> mx = nodes[0].rcoord;
    for (size_t i = 1; i < nodes.size(); i++)
        for (int j = 0; j < 3; j++)
        {
            mn[j] = min(mn[j], nodes[i].rcoord[j]);
            mx[j] = max(mx[j], nodes[i].rcoord[j]);
        }
    valarray<double> scl = mx - mn;
    for (size_t i = 0; i < nodes.size(); i++)
        for (int j = 0; j < 3; j++)
            nodes[i].coord[j] = (int) ((nodes[i].rcoord[j] - mn[j]) / (scl[j]) * ((1 << bits) - 1));
}

static int metric(const valarray<int> &x)
{
    if (max_metric)
        return max(x.max(), -x.min());
    else
        return (x * x).sum();
}

static int mst_metric(int cur, int nxt)
{
    valarray<int> delta = nodes[nxt].coord - nodes[cur].coord;

    if (slew != 0 && parent[cur] >= 0)
    {
        valarray<int> delta2 = nodes[cur].coord - nodes[parent[cur]].coord;
        delta = delta * slew - delta2;
    }
    return metric(delta);
}

static void mst_full()
{
    size_t cur;
    size_t N = nodes.size();
    vector<bool> done(N, false);
    vector<int> prio(N + 1, INT_MAX);
    parent.resize(N, -1);
    order.resize(N);

    prio[0] = 0;
    for (size_t run = 0; run < N; run++)
    {
        cur = N;
        for (size_t i = 0; i < N; i++)
            if (!done[i] && prio[i] < prio[cur]) cur = i;
        assert(cur != N);

        done[cur] = true;
        order[run] = cur;
        for (size_t i = 0; i < N; i++)
        {
            int p;
            if (done[i]) continue;

            p = mst_metric(cur, i);
            if (p < prio[i])
            {
                prio[i] = p;
                parent[i] = cur;
            }
        }
    }
}

static void mst_mesh()
{
    size_t N = nodes.size();
    priority_queue<pair<int, int>, vector<pair<int, int> >, greater<pair<int, int> > > q;
    vector<bool> done(N, false);
    vector<int> prio(N, INT_MAX);
    parent.resize(N, -1);
    order.resize(N);
    int run = 0;

    prio[0] = 0;
    q.push(make_pair(0, 0));
    while (!q.empty())
    {
        int cur = q.top().second;
        q.pop();
        if (done[cur]) continue;
        done[cur] = true;
        order[run++] = cur;
        for (size_t i = 0; i < nodes[cur].edges.size(); i++)
        {
            int nxt = nodes[cur].edges[i].target;
            int p;
            if (done[nxt]) continue;

            p = mst_metric(cur, nxt);
            if (p < prio[nxt])
            {
                prio[nxt] = p;
                parent[nxt] = cur;
                q.push(make_pair(p, nxt));
            }
        }
    }
}

axial_encoder::axial_encoder(int max_length)
{
    for (int i = 0; i < 3; i++)
        for (int j = -max_length; j <= max_length; j++)
            axes[i].code(j);
}

void axial_encoder::output(compressor &c, const valarray<int> &d)
{
    c.encode(axes[0], d[0]);
    c.encode(axes[1], d[1]);
    c.encode(axes[2], d[2]);
}

static int l2(int x)
{
    int ans = 0;

    x = abs(x);
    while (x)
    {
        x >>= 1;
        ans++;
    }
    return ans;
}

conditional_encoder::conditional_encoder(int max_length, bool lg) : lg(lg)
{
    for (int i = -max_length; i <= max_length; i++)
    {
        xfreq.code(i);
        xyfreq[i] = context();
        xzfreq[i] = context();
        for (int j = -max_length; j <= max_length; j++)
        {
            xyfreq[i].code(j);
            xzfreq[i].code(j);
        }
    }
}

void conditional_encoder::output(compressor &c, const valarray<int> &d)
{
    int key = d[0];

    if (lg) key = l2(key);
    c.encode(xfreq, d[0]);
    c.encode(xyfreq[key], d[1]);
    c.encode(xzfreq[key], d[2]);
}

static flip_vector update(const valarray<int> &delta,
                          const valarray<int> &delta2)
{
    flip_vector out;

    out.ctx = CTX_PLAIN;
    out.value = delta;

    valarray<int> delta3 = delta - delta2;
    if (flipflop == ALWAYS
        || (flipflop == MAYBE && metric(delta3) < metric(delta)))
    {
        out.value = delta3;
        out.ctx = CTX_FLIP;
    }
    else if (reorient)
    {
        if (delta2[0] < 0) out.value[0] *= -1;
        if (delta2[1] < 0) out.value[1] *= -1;
        if (delta2[2] < 0) out.value[2] *= -1;

        int old[3] = {abs(delta2[0]), abs(delta2[1]), abs(delta2[2])};
        int nu[3] = {out.value[0], out.value[1], out.value[2]};
        int order[3];

        order[0] = max_element(old, old + 3) - old;
        order[2] = min_element(old, old + 3) - old;
        if (order[2] == order[0]) order[2]++;
        order[1] = 3 - order[0] - order[2];
        out.value[0] = nu[order[0]];
        out.value[1] = nu[order[1]];
        out.value[2] = nu[order[2]];
        if (abs(out.value[1] > 10000))
            cout << out.value << "\n";
    }
    return out;
}

static valarray<int> predict_simple(int x)
{
    if (parent[x] < 0) return valarray<int>(0, 3);
    else return nodes[x].coord - nodes[parent[x]].coord;
}

static inline valarray<int> get_delta(int x)
{
    return nodes[x].coord - nodes[parent[x]].coord;
}

template<class T>
static bool is_zero(const valarray<T> &x)
{
    return x.min() == 0.0 && x.max() == 0.0;
}

template<class T, class S>
static double dot(const valarray<T> &x, const valarray<S> &y)
{
    return x[0] * y[0] + x[1] * y[1] + x[2] * y[2];
}

template<class T, class S>
static valarray<double> cross(const valarray<T> &x, const valarray<S> &y)
{
    valarray<double> ans(3);

    ans[0] = x[1] * y[2] - x[2] * y[1];
    ans[1] = x[2] * y[0] - x[0] * y[2];
    ans[2] = x[0] * y[1] - x[1] * y[0];
    return ans;
}

static valarray<int> predict_basis(int x)
{
    valarray<int> d(3), ui(3), ans(3);
    valarray<double> u(3), v(3), w(3);

    if (parent[x] < 0) return nodes[x].coord;
    if (parent[parent[x]] < 0) return get_delta(x);
    ui = get_delta(parent[x]);
    if (is_zero(ui)) return get_delta(x);
    w = normal[parent[x]];
    if (is_zero(w))
    {
        int c[3] = {abs(ui[0]), abs(ui[1]), abs(ui[2])};
        w[min_element(c, c + 3) - c] = 1.0;
    }
    u[0] = ui[0]; u[1] = ui[1]; u[2] = ui[2];
    v = cross(w, u);
    w = cross(u, v);
    u /= sqrt((u * u).sum() / 3.0);
    v /= sqrt((v * v).sum() / 3.0);
    w /= sqrt((w * w).sum() / 3.0);

#if 0
    double scale = 1.0;
    for (int p = -1; p <= 1; p++)
        for (int q = -1; q <= 1; q++)
            for (int r = -1; r <= 1; r++)
            {
                valarray<int> k(3);
                k[0] = p;
                k[1] = q;
                k[2] = r;
                if (is_zero(k)) continue;
                double s[3] = { abs(dot(k, u)), abs(dot(k, v)), abs(dot(k, w)) };
                scale = min(scale, *max_element(s, s + 3));
            }
    u /= scale;
    v /= scale;
    w /= scale;
#endif

    d = get_delta(x);
    ans[0] = (int) round(dot(d, u));
    ans[1] = (int) round(dot(d, v));
    ans[2] = (int) round(dot(d, w));
    v = cross(w, d);
    w = cross(d, v);
    w /= sqrt((w * w).sum());
    normal[x] = w;
    return ans;
}

static inline valarray<int> predict(int cur)
{
    return basis ? predict_basis(cur) : predict_simple(cur);
}

static void execute()
{
    compressor c;

    size_t N = nodes.size();
    context ctx;
    context degree;
    vector<vector<int> > children(N, vector<int>());
    vector<flip_vector> cooked(N);
    encoder *ec[CTX_COUNT];
    int max_length = 0, max_deg = 0;

    normal.resize(nodes.size(), valarray<double>(0.0, 3));

    for (size_t i = 1; i < N; i++)
        children[parent[i]].push_back(i);
    for (size_t i = 1; i < N; i++)
    {
        int cur = order[i];

        valarray<int> delta = predict(cur);
        valarray<int> delta2(0, 3);
        if (parent[cur] != 0)
            delta2 = predict(parent[cur]);
        cooked[cur] = update(delta, delta2);
        int l2 = (cooked[cur].value * cooked[cur].value).sum();
        max_length = max(max_length, (int) ceil(sqrt((double) l2)));
        max_deg = max((size_t) max_deg, children[i].size());
    }

    if (use_conditional)
    {
        ec[CTX_PLAIN] = new conditional_encoder(max_length, true);
        ec[CTX_FLIP] = new conditional_encoder(max_length, true);
    }
    else
    {
        ec[CTX_PLAIN] = new axial_encoder(max_length);
        ec[CTX_FLIP] = new axial_encoder(max_length);
    }
    for (int i = 0; i < CTX_COUNT; i++) ctx.code(i);
    for (int i = 0; i <= max_deg; i++) degree.code(i);

    /* FIXME: depth first */
    for (size_t i = 1; i < N; i++)
    {
        c.encode(ctx, cooked[i].ctx);
        ec[cooked[i].ctx]->output(c, cooked[i].value);
        c.encode(degree, children[i].size());
    }

    for (int i = 0; i < CTX_COUNT; i++)
        delete ec[i];

    cout << "context bits: " << ctx.bits() / (double) (N - 1) << "\n";
    cout << "context entropy: " << ctx.entropy() << "\n";
    cout << "valence bits: " << degree.bits() / (double) (N - 1) << "\n";
    cout << "valence entropy: " << degree.entropy() << "\n";
    cout << c.bits() / (double) (N - 1) << "\n";
}

static void dump(const char *fname)
{
    ofstream out(fname);

    for (size_t i = 1; i < nodes.size(); i++)
    {
        valarray<double> d = nodes[parent[i]].rcoord - nodes[i].rcoord;

        out << nodes[i].rcoord << "\n"
            << nodes[parent[i]].rcoord << "\n"
            << nodes[i].rcoord << "\n";
    }
    out << "\n";
}

static bool optarg_bool()
{
    if (!optarg) return true;
    else if (strcmp(optarg, "yes") == 0) return true;
    else if (strcmp(optarg, "no") == 0) return false;
    else
    {
        cerr << "Please specify 'yes' or 'no'" << "\n";
        exit(1);
    }
}

static choice optarg_choice()
{
    if (!optarg) return ALWAYS;
    else if (strcmp(optarg, "yes") == 0) return ALWAYS;
    else if (strcmp(optarg, "no") == 0) return NEVER;
    else if (strcmp(optarg, "maybe") == 0) return MAYBE;
    else
    {
        cerr << "Please specify 'yes', 'no' or 'maybe'" << "\n";
        exit(1);
    }
}

static void process_args(int *argc, char ***argv)
{
    int c;

    while ((c = getopt(*argc, *argv, "d:f:b:r:m:M:s:c:")) != -1)
        switch (c)
        {
        case 'd':
            dump_file = optarg;
            break;
        case 'f':
            flipflop = optarg_choice();
            break;
        case 'b':
            basis = optarg_bool();
            break;
        case 'r':
            reorient = optarg_bool();
            break;
        case 'm':
            full_mst = optarg_bool();
            break;
        case 'M':
            max_metric = optarg_bool();
            break;
        case 's':
            slew = atoi(optarg);
            break;
        case 'c':
            use_conditional = optarg_bool();
            break;
        case ':':
        case '?':
            exit(0);
        }
    *argc -= optind;
    *argv += optind;
}

int main(int argc, char **argv)
{
    process_args(&argc, &argv);
    assert(argc >= 2);
    readin(argv[0]);
    quantise(atoi(argv[1]));
    if (full_mst) mst_full(); else mst_mesh();
    if (dump_file != "") dump(dump_file.c_str());
    execute();
}
