#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <queue>
#include <stack>
#include <map>
#include <algorithm>
#include <valarray>
#include <cstdlib>
#include <climits>
#include <cmath>
#include <cassert>
#include <cfloat>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>
#include "arith.h"
#include "ply.h"
#include "coders3d.h"
#include "ANN/ANN.h"

enum predictor_t
{
    PRED_PLAIN = 0,
    PRED_LEFT = 1,
    PRED_RIGHT = 2,
    PRED_FORWARD = 3,
    PRED_COUNT = 4
};

enum coder_t
{
    CODER_AXIAL,
    CODER_RADIAL,
    CODER_SINGLE
};

using namespace std;

static int slewn = -1;
static int slewd = 1;
static bool full_mst = true;
static int predictor_set = 0xf;
static coder_t coder_mode = CODER_SINGLE;
static string dump_file = "";
static string ps_file = "";
static string blender_file = "";
static double ps_zoom_from[4] = {DBL_MIN, DBL_MIN, DBL_MIN, DBL_MIN};
static double ps_zoom_to[4] = {DBL_MIN, DBL_MIN, DBL_MIN, DBL_MIN};
static bool compressing = true;

const char * const predictor_names[PRED_COUNT + 1] =
{
    "unbiased",
    "left",
    "right",
    "forward",
    "terminate"
};

const float predictor_vectors[PRED_COUNT][3] =
{
    {0, 0, 0},
    {0, -1, 0},
    {0, 1, 0},
    {1, 0, 0},
};

struct edge
{
    int target;

    edge() {}
    explicit edge(int t) : target(t) {}
};

struct flip_vector
{
    valarray<int> value;
    valarray<float> up;
    predictor_t predictor;

    flip_vector() : value(3), up(3), predictor(PRED_PLAIN) {}
};

struct node
{
    valarray<float> rcoord;
    valarray<int> coord;
    vector<edge> edges;
    int parent;
    int discovery;
    vector<int> children;
    double vertex_bits;
    flip_vector cooked;

    node() : rcoord(3), coord(3), edges(), parent(-1), discovery(-1),
        children(), vertex_bits(0.0), cooked() {}
};

struct cluster_data
{
    vector<node> nodes;
    vector<valarray<int> > faces;
    vector<int> order;
    vector<int> mst_order;
    double mst_max;
    float scale;
    int qbits;
    valarray<float> corner;
};

template<class T>
static ostream &operator <<(ostream &o, const valarray<T> d)
{
    o << d[0] << " " << d[1] << " " << d[2];
    return o;
}

template<class T, class S>
static T dot(const valarray<T> &x, const valarray<S> &y)
{
    return x[0] * y[0] + x[1] * y[1] + x[2] * y[2];
}

static inline double length(const valarray<int> &x)
{
    return sqrt((double) x[0] * (double) x[0]
                + (double) x[1] * (double) x[1]
                + (double) x[2] * (double) x[2]);
}

static inline float length(const valarray<float> &x)
{
    return sqrtf(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
}

static inline int round_length(const valarray<int> &x)
{
    return (int) round(length(x));
}

static inline int ceil_length(const valarray<int> &x)
{
    return (int) ceil(length(x));
}

static bool is_zero(const valarray<int> &x)
{
    return x.min() == 0 && x.max() == 0;
}

static bool is_zero(const valarray<float> &x)
{
    return length(x) == 0.0f;
}

static int metric(const valarray<int> &x)
{
    return dot(x, x);
}

static inline valarray<int> get_delta(int x, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    if (nodes[x].parent < 0) return valarray<int>(0, 3);
    else return nodes[x].coord - nodes[data.nodes[x].parent].coord;
}

static void normalise(valarray<float> &v)
{
    if (!is_zero(v)) v /= length(v);
}

static valarray<float> rotate_normal(const valarray<int> &oldf,
                                     const valarray<int> &newf,
                                     const valarray<float> &oldu)
{
    valarray<float> of(3), nf(3), newu(3);

    of[0] = oldf[0]; of[1] = oldf[1]; of[2] = oldf[2];
    nf[0] = newf[0]; nf[1] = newf[1]; nf[2] = newf[2];
    if (is_zero(of) || is_zero(nf)) return oldu;
    normalise(of);
    normalise(nf);

    if (dot(of, nf) * dot(of, nf) < 0.75)
    {
        newu[0] = of[1] * nf[2] - of[2] * nf[1];
        newu[1] = of[2] * nf[0] - of[0] * nf[2];
        newu[2] = of[0] * nf[1] - of[1] * nf[0];
    }
    else
        newu = oldu - dot(oldu, nf) * nf;
    normalise(newu);
    return newu;
}

/* Returns a prediction using the given predictor, returning the
 * predicted point relative to the parent.
 */
static valarray<int> make_prediction(const valarray<int> &old_forward,
                                     const valarray<float> &old_up,
                                     predictor_t predictor)
{
    float l = sqrtf((float) dot(old_forward, old_forward));
    valarray<float> old_left(0.0f, 3);
    old_left[0] = old_up[1] * old_forward[2] - old_up[2] * old_forward[1];
    old_left[1] = old_up[2] * old_forward[0] - old_up[0] * old_forward[2];
    old_left[2] = old_up[0] * old_forward[1] - old_up[1] * old_forward[0];
    normalise(old_left);

    valarray<float> pred(3);
    valarray<int> p(3);

    for (int j = 0; j < 3; j++)
        pred[j] = predictor_vectors[predictor][0] * old_forward[j]
            + predictor_vectors[predictor][1] * old_left[j]
            + predictor_vectors[predictor][2] * old_up[j];
    if (is_zero(pred))
        return valarray<int>(0, 3);
    for (int j = 0; j < 3; j++)
        p[j] = (int) round(l * pred[j] / length(pred));
    return p;
}

static flip_vector update(const valarray<int> &delta,
                          const valarray<int> &old_forward,
                          const valarray<float> &old_up)
{
    flip_vector out;
    int best_metric = INT_MAX;

    if (!is_zero(delta))
        out.up = rotate_normal(old_forward, delta, old_up);
    else
        out.up = old_up;

    for (int i = 0; i < (int) PRED_COUNT; i++)
    {
        valarray<int> p(0, 3);

        if (!(predictor_set & (1 << i))) continue;
        p = delta - make_prediction(old_forward, old_up, (predictor_t) i);
        if (metric(p) < best_metric)
        {
            out.value = p;
            out.predictor = (predictor_t) i;
            best_metric = metric(p);
        }
    }

    return out;
}

struct mst_data
{
    priority_queue<pair<float, int>, vector<pair<float, int> >, greater<pair<float, int> > > q;
    vector<bool> done;
    vector<float> prio;
    int run;
    bool fancy;
};

static float mst_metric(int cur, int nxt,
                        const mst_data &mst, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    valarray<int> delta = nodes[nxt].coord - nodes[cur].coord;

    if (slewn < 0 && nodes[cur].parent >= 0)
    {
        valarray<int> delta2 = nodes[cur].coord - nodes[nodes[cur].parent].coord;
        float m = length(delta2 - delta);
        float d = log(m + 1.0) * length(delta);
        return d;
    }
    else if (slewn != 0 && nodes[cur].parent >= 0)
    {
        valarray<int> delta2 = nodes[cur].coord - nodes[nodes[cur].parent].coord;
        float x = dot(delta, delta2) / length(delta2);
        float y = sqrt(max(0.0f, dot(delta, delta) - x * x));
        if (x < 0.0f) return length(delta);
        else return sqrtf(x * x * slewn / slewd + y * y);
    }
    else
        return length(delta);
}

static void mst_update(cluster_data &data, mst_data &mst, int cur)
{
    vector<node> &nodes = data.nodes;

    mst.done[cur] = true;
    nodes[cur].discovery = mst.run;
    data.order[mst.run++] = cur;
    valarray<int> delta = get_delta(cur, data);
    valarray<int> old(0, 3);
    valarray<float> up(0.0f, 3);
    up[2] = -1.0;
    if (nodes[cur].parent != -1)
    {
        old = get_delta(nodes[cur].parent, data);
        up = nodes[nodes[cur].parent].cooked.up;
    }
    assert(length(up) > 1e-10);
    nodes[cur].cooked = update(delta, old, up);

    for (size_t i = 0; i < nodes[cur].edges.size(); i++)
    {
        int nxt = nodes[cur].edges[i].target;
        float p;
        if (mst.done[nxt]) continue;

        p = mst.fancy ? mst_metric(cur, nxt, mst, data)
            : metric(nodes[nxt].coord - nodes[cur].coord);
        if (p < mst.prio[nxt])
        {
            mst.prio[nxt] = p;
            nodes[nxt].parent = cur;
            mst.q.push(make_pair(p, nxt));
        }
    }
}

static bool mst_mesh(cluster_data &data, bool fancy)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    mst_data mst;
    mst.done.resize(N, false);
    mst.prio.resize(N, INT_MAX);
    mst.fancy = fancy;
    mst.run = 0;
    data.order.resize(N);

    mst.prio[0] = 0;
    mst.q.push(make_pair(0, 0));
    while (!mst.q.empty())
    {
        int cur = mst.q.top().second;
        mst.q.pop();
        if (mst.done[cur]) continue;
        mst_update(data, mst, cur);
        if (fancy && cur != 0)
        {
            valarray<int> last = get_delta(cur, data);
            do
            {
                int nxt = -1;
                float nxt_metric = FLT_MAX;
                for (size_t i = 0; i < nodes[cur].edges.size(); i++)
                {
                    int trg = nodes[cur].edges[i].target;
                    if (mst.done[trg]) continue;
                    float trg_metric = mst_metric(cur, trg, mst, data);
                    if (trg_metric < nxt_metric)
                    {
                        nxt = trg;
                        nxt_metric = trg_metric;
                    }
                }
                if (nxt != -1)
                {
                    nodes[nxt].parent = cur;
                    mst_update(data, mst, nxt);
                }
                cur = nxt;
            } while (cur != -1);
        }
    }
    return mst.run == (int) N;
}

static void create_edges(cluster_data &data, ANNpointSet *t, int high)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 0; i < N; i++)
    {
        ANNcoord q[3];
        q[0] = nodes[i].coord[0];
        q[1] = nodes[i].coord[1];
        q[2] = nodes[i].coord[2];
        int k = t->annkFRSearch(q, high * high + 0.5);
        ANNidxArray nn_idx = new ANNidx[k];
        ANNdistArray dd = new ANNdist[k];
        t->annkFRSearch(q, high * high + 0.5, k, nn_idx, dd);

        nodes[i].edges.clear();
        for (int j = 0; j < k; j++)
        {
            int id = nn_idx[j];
            nodes[i].edges.push_back(edge(id));
        }

        delete[] nn_idx;
        delete[] dd;
    }
}

static bool connected(const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    vector<bool> done(N, false);
    vector<int> st;

    st.reserve(N);
    st.push_back(0);
    N--;
    while (!st.empty())
    {
        int cur = st.back();
        st.pop_back();
        done[cur] = true;
        for (size_t i = 0; i < nodes[cur].edges.size(); i++)
        {
            int trg = nodes[cur].edges[i].target;
            if (!done[trg])
            {
                done[trg] = true;
                st.push_back(trg);
                N--;
            }
        }
    }
    return N == 0;
}

static void mst_full(cluster_data &data)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    ANNcoord *points = new ANNcoord[3 * N];
    ANNpoint *point_array = new ANNpoint[N];
    ANNpointSet *t;
    int lmax = 1;

    for (size_t i = 0; i < N; i++)
    {
        points[i * 3 + 0] = nodes[i].coord[0];
        points[i * 3 + 1] = nodes[i].coord[1];
        points[i * 3 + 2] = nodes[i].coord[2];
        point_array[i] = points + i * 3;
    }
    t = new ANNkd_tree(point_array, N, 3);

    do
    {
        lmax *= 2;
        create_edges(data, t, lmax);
    } while (!connected(data));
    mst_mesh(data, false);

    data.mst_max = 0;
    for (size_t i = 0; i < N; i++)
    {
        valarray<int> delta = get_delta(i, data);
        data.mst_max = max(data.mst_max, sqrt((double) dot(delta, delta)));
    }
    lmax = (int) ceil(data.mst_max);
    create_edges(data, t, lmax);
    mst_mesh(data, true);

    delete t;
    delete[] points;
    delete[] point_array;
}

struct coder_data
{
    int max_length;
    int max_x;
    int max_y;
    int max_z;
    coder *geometry;
    context predictor_context;

    coder_data() : max_length(0), max_x(0), max_y(0), max_z(0), geometry(NULL), predictor_context() {}
};

template<class T>
static void show_context(const char *name, const T &ctx)
{
    double itotal = 1.0 / ctx.use();
    cout << name << ": "
        << ctx.bits() * itotal << " | "
        << ctx.entropy() * itotal << " ("
        << ctx.use() << ")\n";
}

class predictor_comparator : public binary_function<int, int, bool>
{
private:
    const cluster_data &data;

public:
    predictor_comparator(const cluster_data &d) : data(d) {}

    bool operator()(int a, int b)
    {
        return data.nodes[a].cooked.predictor < data.nodes[b].cooked.predictor;
    }
};

/* Reorders vertices based on a depth first walk of the tree, visiting
 * children in the order determined by the predictor.
 */
static void reorder(cluster_data &data)
{
    vector<node> &nodes = data.nodes;
    stack<int> s;
    int index = 0;

    s.push(0);
    while (!s.empty())
    {
        int cur = s.top();
        s.pop();
        data.order[index++] = cur;
        sort(nodes[cur].children.begin(), nodes[cur].children.end(), predictor_comparator(data));
        for (vector<int>::reverse_iterator i = nodes[cur].children.rbegin();
             i != nodes[cur].children.rend(); i++)
            s.push(*i);
    }
}

static void cook(cluster_data &data, coder_data *coders)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 1; i < N; i++)
    {
        nodes[nodes[i].parent].children.push_back(i);
        int c = nodes[i].cooked.predictor;
        coders[c].max_length = max(coders[c].max_length, ceil_length(nodes[i].cooked.value));
        coders[c].max_x = max(coders[c].max_x, abs(nodes[i].cooked.value[0]));
        coders[c].max_y = max(coders[c].max_y, abs(nodes[i].cooked.value[1]));
        coders[c].max_z = max(coders[c].max_z, abs(nodes[i].cooked.value[2]));
    }

    data.mst_order = data.order; /* Save for the Blender dumper */
    reorder(data);
}

static void prepare_encoders(coder_data *coders)
{
    for (int i = 0; i < (int) PRED_COUNT; i++)
    {
        switch (coder_mode)
        {
        case CODER_AXIAL:
            coders[i].geometry = new axial_coder(coders[i].max_x,
                                                 coders[i].max_y,
                                                 coders[i].max_z);
            break;
        case CODER_RADIAL:
            coders[i].geometry = new radial_coder(coders[i].max_length);
            break;
        case CODER_SINGLE:
            coders[i].geometry = new single_coder(coders[i].max_x,
                                                  coders[i].max_y,
                                                  coders[i].max_z);
            break;
        }
    }

    for (int i = 0; i < (int) PRED_COUNT; i++)
        coders[i].predictor_context.code_range(i, PRED_COUNT);
    coders[PRED_COUNT].predictor_context.code_range(0, PRED_COUNT);
}

static void compress(cluster_data &data, compressor &c, coder_data *coders)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 0; i < N; i++)
    {
        int cur = data.order[i];

        // Encode position. The parent provides the predictor
        if (cur != 0)
        {
            int predictor = nodes[cur].cooked.predictor;
            nodes[cur].vertex_bits += coders[predictor].geometry->encode(c, nodes[cur].cooked.value);
        }

        // Encode valence and child predictors
        predictor_t last = PRED_COUNT;
        for (size_t j = 0; j < nodes[cur].children.size(); j++)
        {
            int nxt = nodes[cur].children[j];
            nodes[nxt].vertex_bits += c.encode(coders[last].predictor_context, nodes[nxt].cooked.predictor);
            last = nodes[nxt].cooked.predictor;
        }
        nodes[cur].vertex_bits += c.encode(coders[last].predictor_context, PRED_COUNT);
    }
}

static void predictor_stats(predictor_t predictor, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    size_t count = 0;
    double l2 = 0.0;

    for (size_t j = 1; j < N; j++)
    {
        if (predictor == PRED_COUNT
            || predictor == nodes[j].cooked.predictor)
        {
            count++;
            l2 += dot(nodes[j].cooked.value, nodes[j].cooked.value);
        }
    }
    cout << "    count: " << count << "\n";
    if (count)
        cout << "    L2: " << sqrt(l2 / count) << "\n";
    else
        cout << "    L2: -1\n";
}

static void statistics(compressor &c, const cluster_data &data, coder_data *coders)
{
    size_t points = data.nodes.size();
    double context_bits = 0.0, context_entropy = 0.0;
    for (int i = 0; i <= (int) PRED_COUNT; i++)
    {
        string title = string("predictor (") + predictor_names[i] + ")";
        show_context(title.c_str(), coders[i].predictor_context);
        context_bits += coders[i].predictor_context.bits();
        context_entropy += coders[i].predictor_context.entropy();
    }
    for (int i = 0; i < (int) PRED_COUNT; i++)
    {
        cout << predictor_names[i] << "\n"
            << "    max length: " << coders[i].max_length << "\n";
        show_context("    geometry", *coders[i].geometry);
        predictor_stats((predictor_t) i, data);
    }
    cout << "total\n";
    predictor_stats(PRED_COUNT, data);
    cout << "points: " << points << "\n";
    cout << "tree: "
        << context_bits / points << " | "
        << context_entropy / points << "\n";
    cout << "geometry: "
        << (c.bits() - context_bits) / points << "\n";
    cout << "total: " << c.bits() / (double) points << "\n";
    cout << "\n";
    // coders[PRED_FORWARD].geometry->dump(cout);
}

static void dump(const string &fname, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    ofstream out(fname.c_str());

    out << "ply\n"
        << "format ascii 1.0\n"
        << "element vertex " << 2 * N << "\n"
        << "property float32 x\n"
        << "property float32 y\n"
        << "property float32 z\n"
        << "element face " << N - 1 << "\n"
        << "property list uint8 int32 vertex_indices\n"
        << "end_header\n";

    for (size_t i = 0; i < N; i++)
    {
        valarray<float> ofs1(3), ofs2(3);
        for (int j = 0; j < 3; j++)
        {
            ofs1[j] = nodes[i].rcoord[j] + 20.0f * nodes[i].cooked.up[j] / data.scale;
            ofs2[j] = nodes[i].rcoord[j] - 20.0f * nodes[i].cooked.up[j] / data.scale;
        }
        out << ofs1 << "\n" << ofs2 << "\n";
    }
    for (size_t i = 1; i < N; i++)
    {
        int p = nodes[i].parent;
        out << "4 " << 2 * i << " " << 2 * i + 1 << " "
            << 2 * p + 1 << " " << 2 * p << "\n";
    }
}

struct element_comparator : public binary_function<vector<int>, vector<int>, bool>
{
private:
    const cluster_data &data;

public:
    element_comparator(const cluster_data &_data) : data(_data) {}
    bool operator()(const vector<int> &a, const vector<int> &b) const;
};

bool element_comparator::operator()(const vector<int> &a, const vector<int> &b) const
{
    double za = 0.0, zb = 0.0;

    for (size_t i = 0; i < a.size(); i++)
        za += data.nodes[a[i]].coord[2];
    za /= a.size();
    if (a.size() < 3) za += 32.0 * (3 - a.size());
    for (size_t i = 0; i < b.size(); i++)
        zb += data.nodes[b[i]].coord[2];
    zb /= b.size();
    if (b.size() < 3) zb += 32.0 * (3 - b.size());
    return za < zb;
}

static void dump_ps_element(ostream &out, const vector<int> &element,
                            double *bbox, const cluster_data &data)
{
    if (bbox)
    {
        int x1 = INT_MAX, y1 = INT_MAX, x2 = INT_MIN, y2 = INT_MIN;

        for (size_t i = 0; i < element.size(); i++)
        {
            const valarray<int> &coord = data.nodes[element[i]].coord;
            x1 = min(x1, coord[0]);
            y1 = min(y1, coord[1]);
            x2 = max(x2, coord[0]);
            y2 = max(y2, coord[1]);
        }

        if (x1 > bbox[2] + 32
            || y1 > bbox[3] + 32
            || x2 < bbox[0] - 32
            || y2 < bbox[1] - 32) return;
    }

    for (size_t i = 0; i < element.size(); i++)
        out << data.nodes[element[i]].coord[0] << " "
            << data.nodes[element[i]].coord[1] << " ";

    if (element.size() >= 3) out << element.size() << " polygon\n";
    else if (element.size() == 1) out << "vertex\n";
    else out << predictor_names[data.nodes[element[0]].cooked.predictor] << "\n";
}

static void dump_ps(const string &fname, const cluster_data &data)
{
    vector<vector<int> > elements;
    const vector<node> &nodes = data.nodes;
    ofstream out(fname.c_str());

    elements.reserve(data.faces.size() + 2 * nodes.size());
    for (size_t i = 0; i < data.faces.size(); i++)
    {
        elements.push_back(vector<int>(data.faces[i].size()));
        for (size_t j = 0; j < data.faces[i].size(); j++)
            elements.back()[j] = data.faces[i][j];
    }
    for (size_t i = 1; i < nodes.size(); i++)
    {
        vector<int> edge(2);
        edge[0] = i;
        edge[1] = nodes[i].parent;
        elements.push_back(edge);
    }
    for (size_t i = 0; i < nodes.size(); i++)
        elements.push_back(vector<int>(1, i));

    sort(elements.begin(), elements.end(), element_comparator(data));
    double size = 512.0;
    double scale = size / (1 << data.qbits);
    out << "%!PS-Adobe-2.0 EPSF-1.2\n";
    out << "%%BoundingBox: 0.0 0.0 " << size << " " << size << "\n";
    out << "%%EndComments\n";
    out <<
        "gsave\n"
        "/drawedge { newpath moveto lineto stroke } def\n"
        "/forward { 1.0 0.0 0.0 setrgbcolor drawedge } bind def\n"
        "/left { 0.0 1.0 0.0 setrgbcolor drawedge } bind def\n"
        "/right { 0.0 0.0 1.0 setrgbcolor drawedge } bind def\n"
        "/unbiased { 0.9 0.9 0.0 setrgbcolor drawedge } bind def\n"
        "/polygon { 0.9 setgray newpath 3 1 roll moveto 1 sub { lineto } repeat closepath fill } bind def\n"
        "/vertex { 0.0 setgray newpath 1 index 1 index moveto 4 0 360 arc closepath fill } bind def\n"
        "/rectangle { newpath\n"
        "             3 index 3 index moveto\n"
        "             3 index 1 index lineto\n"
        "             1 index 1 index lineto\n"
        "             1 index 3 index lineto\n"
        "             closepath pop pop pop pop } def\n";
    out << "gsave\n"
        << scale << " " << scale << " scale\n"
        << "/Times findfont 4 scalefont setfont\n";
    for (size_t i = 0; i < elements.size(); i++)
        dump_ps_element(out, elements[i], NULL, data);
    if (ps_zoom_from[0] != DBL_MIN && ps_zoom_to[0] != DBL_MIN)
    {
        out << "/zoomfrom { "
            << ps_zoom_from[0] << " " << ps_zoom_from[1] << " "
            << ps_zoom_from[2] << " " << ps_zoom_from[3] << " } def\n";
        out << "/zoomto { "
            << ps_zoom_to[0] << " " << ps_zoom_to[1] << " "
            << ps_zoom_to[2] << " " << ps_zoom_to[3] << " } def\n";

        out << "gsave\n";
        out << "zoomto rectangle clip\n"
            << ps_zoom_to[0] << " " << ps_zoom_to[1] << " translate\n"
            << (ps_zoom_to[2] - ps_zoom_to[0]) / (ps_zoom_from[2] - ps_zoom_from[0]) << " "
            << (ps_zoom_to[3] - ps_zoom_to[1]) / (ps_zoom_from[3] - ps_zoom_from[1]) << " scale\n"
            << -ps_zoom_from[0] << " " << -ps_zoom_from[1] << " translate\n";
        for (size_t i = 0; i < elements.size(); i++)
            dump_ps_element(out, elements[i], ps_zoom_from, data);
        out << "grestore\n";

        out << "gsave 0.0 setgray 8 setlinewidth\n"
            << "zoomfrom rectangle stroke\n"
            << "zoomto rectangle stroke\n"
            << "grestore\n";
    }
    out << "grestore\n";
}

static void dump_blender(const string &fname, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    ofstream out(fname.c_str());

    for (size_t i = 0; i < N; i++)
        out << nodes[i].rcoord << "\n";
    for (size_t i = 1; i < N; i++)
    {
        int y = data.mst_order[i];
        int x = nodes[y].parent;
        out << nodes[x].rcoord << " " << nodes[y].rcoord << " "
            << nodes[y].cooked.predictor << "\n";
    }
}

static bool optarg_bool()
{
    if (!optarg) return true;
    else if (strcmp(optarg, "yes") == 0) return true;
    else if (strcmp(optarg, "no") == 0) return false;
    else
    {
        cerr << "Please specify 'yes' or 'no'\n";
        exit(1);
    }
}

static vector<string> optarg_commas()
{
    char *cur, *nxt;
    vector<string> ans;

    cur = optarg;
    while (*cur)
    {
        while (*cur == ',') cur++;
        if (*cur == '\0') break;
        nxt = cur;
        while (*nxt != ',' && *nxt != '\0') nxt++;
        ans.push_back(string(cur, nxt - cur));
        cur = nxt;
    }
    return ans;
}

static int optarg_predictor_set()
{
    vector<string> predictors = optarg_commas();
    int ps = 0;

    for (size_t i = 0; i < predictors.size(); i++)
    {
        int j;
        for (j = 0; j < PRED_COUNT; j++)
            if (predictors[i] == predictor_names[i]) break;
        if (j == PRED_COUNT)
        {
            cerr << "Please specify a set of predictors, separated by commas\n";
            exit(1);
        }
        ps |= 1 << j;
    }
    return ps;
}

static void optarg_zoom(double *zoom)
{
    vector<string> fields = optarg_commas();

    if (fields.size() != 4)
    {
        cerr << "Please specify zoom location in form x1,y1,x2,y2\n";
        exit(1);
    }
    zoom[0] = atof(fields[0].c_str());
    zoom[1] = atof(fields[1].c_str());
    zoom[2] = atof(fields[2].c_str());
    zoom[3] = atof(fields[3].c_str());
}

static void process_args(int *argc, char ***argv)
{
    int c;
    enum longopt_t
    {
        ZOOM_FROM = 256,
        ZOOM_TO
    };
    const struct option longopts[] =
    {
        { "blender", 1, NULL, 'b' },
        { "dump", 1, NULL, 'd' },
        { "predictors", 1, NULL, 'p' },
        { "mst", 1, NULL, 'm' },
        { "slewn", 1, NULL, 's' },
        { "slewd", 1, NULL, 'S' },
        { "encoder", 1, NULL, 'e' },
        { "ps", 1, NULL, 'P' },
        { "decompress", 1, NULL, 'D' },
        { "zoom-from", 1, NULL, (int) ZOOM_FROM },
        { "zoom-to", 1, NULL, (int) ZOOM_TO },
        { NULL, 0, NULL, 0 }
    };

    while ((c = getopt_long(*argc, *argv, "d:p:m:s:S:e:P:b:D", longopts, NULL)) != -1)
        switch (c)
        {
        case 'b':
            blender_file = optarg;
            break;
        case 'd':
            dump_file = optarg;
            break;
        case 'P':
            ps_file = optarg;
            break;
        case 'p':
            predictor_set = optarg_predictor_set();
            break;
        case 'm':
            full_mst = optarg_bool();
            break;
        case 's':
            slewn = atoi(optarg);
            break;
        case 'S':
            slewd = atoi(optarg);
            break;
        case 'e':
            if (strcmp(optarg, "radial") == 0)
                coder_mode = CODER_RADIAL;
            else if (strcmp(optarg, "axial") == 0)
                coder_mode = CODER_AXIAL;
            else if (strcmp(optarg, "single") == 0)
                coder_mode = CODER_SINGLE;
            else
            {
                cerr << "No such encoder " << optarg << "\n";
                exit(1);
            }
            break;
        case 'D': compressing = false; break;
        case ZOOM_FROM:
            optarg_zoom(ps_zoom_from);
            break;
        case ZOOM_TO:
            optarg_zoom(ps_zoom_to);
            break;
        case ':':
        case '?':
            exit(0);
        }
    *argc -= optind;
    *argv += optind;
}

static void readin(const char *fname, cluster_data &data)
{
    vector<valarray<float> > vertices;
    vector<valarray<int> > &faces = data.faces;

    ifstream f(fname);
    if (!f)
    {
        cerr << "Cannot open " << fname << "\n";
        exit(1);
    }
    read_ply(f, vertices, faces);

    data.nodes.resize(vertices.size());
    for (size_t i = 0; i < vertices.size(); i++)
        data.nodes[i].rcoord = vertices[i];

    for (size_t i = 0; i < faces.size(); i++)
    {
        for (size_t j = 0; j < faces[i].size(); j++)
        {
            int p = faces[i][j];
            int q = faces[i][(j + 1) % faces[i].size()];
            data.nodes[p].edges.push_back(edge(q));
            data.nodes[q].edges.push_back(edge(p));
        }
    }
}

static void quantise(int bits, cluster_data &data)
{
    data.qbits = bits;
    valarray<float> mn = data.nodes[0].rcoord;
    valarray<float> mx = data.nodes[0].rcoord;
    for (size_t i = 1; i < data.nodes.size(); i++)
        for (int j = 0; j < 3; j++)
        {
            mn[j] = min(mn[j], data.nodes[i].rcoord[j]);
            mx[j] = max(mx[j], data.nodes[i].rcoord[j]);
        }
    float scale = ((1 << bits) - 1) / (mx - mn).max();
    for (size_t i = 0; i < data.nodes.size(); i++)
        for (int j = 0; j < 3; j++)
            data.nodes[i].coord[j] = (int) floor((data.nodes[i].rcoord[j] - mn[j]) * scale);
    data.scale = scale;
    data.corner.resize(3);
    data.corner = mn;

    size_t bottom = 0;
    for (size_t i = 1; i < data.nodes.size(); i++)
    {
        bool smaller = false;
        for (int j = 2; j >= 0; j--)
            if (data.nodes[i].coord[j] != data.nodes[bottom].coord[j])
            {
                smaller = data.nodes[i].coord[j] < data.nodes[bottom].coord[j];
                break;
            }
        if (smaller) bottom = i;
    }

    swap(data.nodes[0], data.nodes[bottom]);
    for (size_t i = 0; i < data.nodes.size(); i++)
        for (size_t j = 0; j < data.nodes[i].edges.size(); j++)
        {
            if (data.nodes[i].edges[j].target == 0)
                data.nodes[i].edges[j].target = bottom;
            else if (data.nodes[i].edges[j].target == (int) bottom)
                data.nodes[i].edges[j].target = 0;
        }

    for (size_t i = 0; i < data.faces.size(); i++)
        for (size_t j = 0; j < data.faces[i].size(); j++)
            if (data.faces[i][j] == 0)
                data.faces[i][j] = bottom;
            else if (data.faces[i][j] == (int) bottom)
                data.faces[i][j] = 0;
}

static void write_header(ostream &out, const cluster_data &data,
                         const coder_data *coders)
{
    uint32_t N = data.nodes.size();
    uint8_t ps = predictor_set;
    uint8_t cm = coder_mode;
    out.write((const char *) &N, sizeof(N));
    out.write((const char *) &data.scale, sizeof(data.scale));
    for (int i = 0; i < 3; i++)
        out.write((const char *) &data.corner[i], sizeof(data.corner[i]));
    for (int i = 0; i < 3; i++)
    {
        uint32_t w = data.nodes[0].coord[i];
        out.write((const char *) &w, sizeof(w));
    }
    for (int i = 0; i < 3; i++)
        out.write((const char *) &data.nodes[0].cooked.up[i], sizeof(data.nodes[0].cooked.up[i]));
    out.write((const char *) &ps, sizeof(ps));
    out.write((const char *) &cm, sizeof(cm));
    for (int i = 0; i < PRED_COUNT; i++)
        coders[i].geometry->write_header(out);
}

/**********************
 * Decompression code *
 **********************/

static void read_header(istream &in, cluster_data &data,
                        coder_data *coders)
{
    uint32_t N;
    uint8_t ps;
    uint8_t cm;
    uint32_t w;

    in.read((char *) &N, sizeof(N));
    in.read((char *) &data.scale, sizeof(data.scale));
    data.corner.resize(3);
    for (int i = 0; i < 3; i++)
        in.read((char *) &data.corner[i], sizeof(data.corner[i]));
    data.nodes.resize(N);
    data.nodes[0].cooked.up.resize(3);
    for (int i = 0; i < 3; i++)
    {
        in.read((char *) &w, sizeof(w));
        data.nodes[0].coord[i] = w;
    }
    for (int i = 0; i < 3; i++)
        in.read((char *) &data.nodes[0].cooked.up[i], sizeof(data.nodes[0].cooked.up[i]));
    in.read((char *) &ps, sizeof(ps)); predictor_set = ps;
    in.read((char *) &cm, sizeof(cm)); coder_mode = (coder_t) cm;
    for (int i = 0; i < PRED_COUNT; i++)
    {
        switch (cm)
        {
        case CODER_AXIAL:
            coders[i].geometry = new axial_coder(in); break;
        case CODER_RADIAL:
            coders[i].geometry = new radial_coder(in); break;
        case CODER_SINGLE:
            coders[i].geometry = new single_coder(in); break;
        default: abort();
        }
        coders[i].predictor_context.code_range(i, PRED_COUNT);
    }
    coders[PRED_COUNT].predictor_context.code_range(0, PRED_COUNT);
    data.nodes.resize(N);
}

static void decompress(decompressor &d, cluster_data &data,
                       coder_data *coders)
{
    vector<node> &nodes = data.nodes;
    size_t N = 1;
    stack<int> s;

    s.push(0);
    while (!s.empty())
    {
        int cur = s.top();
        s.pop();
        if (cur != 0)
        {
            /* Recover position. The parent provided the predictor */

            int parent = nodes[cur].parent;
            predictor_t predictor = nodes[cur].cooked.predictor;
            valarray<int> forward = get_delta(parent, data);
            valarray<int> base = make_prediction(forward,
                                                 nodes[parent].cooked.up,
                                                 predictor);
            valarray<int> correction = coders[predictor].geometry->decode(d);
            nodes[cur].coord = nodes[parent].coord + base + correction;
            nodes[cur].cooked.up.resize(3);
            nodes[cur].cooked.up = rotate_normal(forward,
                                                 base + correction,
                                                 nodes[parent].cooked.up);
        }

        /* Recover predictors and valence */
        predictor_t last = PRED_COUNT;
        while (1)
        {
            predictor_t predictor = (predictor_t) d.decode(coders[last].predictor_context);
            if (predictor == PRED_COUNT) break; /* terminator */
            nodes[N].cooked.predictor = predictor;
            nodes[cur].children.push_back(N);
            nodes[N].parent = cur;
            N++;
            last = predictor;
        }
        for (vector<int>::reverse_iterator i = nodes[cur].children.rbegin();
             i != nodes[cur].children.rend(); i++)
            s.push(*i);
    }
}

static void dequantise(cluster_data &data)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 0; i < N; i++)
        for (int j = 0; j < 3; j++)
            nodes[i].rcoord[j] = data.corner[j] + nodes[i].coord[j] / data.scale;
}

static void writeout(const char *fname, cluster_data &data)
{
    ofstream out(fname);
    out.exceptions(ios::failbit);

    out << "ply\n"
        << "format ascii 1.0\n"
        << "comment created by smst3d\n"
        << "element vertex " << data.nodes.size() << "\n"
        << "property float32 x\n"
        << "property float32 y\n"
        << "property float32 z\n"
        << "property float32 nx\n"
        << "property float32 ny\n"
        << "property float32 nz\n"
        << "element face 0\n"
        << "property list uint8 int32 vertex_indices\n"
        << "end_header\n";
    for (size_t i = 0; i < data.nodes.size(); i++)
        out << data.nodes[i].rcoord << " " << data.nodes[i].cooked.up << "\n";
    out.close();
}

int main(int argc, char **argv)
{
    coder_data coders[PRED_COUNT + 1];
    cluster_data data;

    process_args(&argc, &argv);

    if (compressing)
    {
        assert(argc == 3);

        compressor c;
        ofstream out(argv[2]);
        out.exceptions(ios::failbit);

        readin(argv[0], data);
        quantise(atoi(argv[1]), data);
        if (full_mst) mst_full(data); else mst_mesh(data, true);
        cook(data, coders);
        prepare_encoders(coders);

        write_header(out, data, coders);
        c.open(out);
        compress(data, c, coders);
        if (dump_file != "") dump(dump_file, data);
        if (ps_file != "") dump_ps(ps_file, data);
        if (blender_file != "") dump_blender(blender_file, data);
        cout << "L: " << data.mst_max << "\n";

        statistics(c, data, coders);
        c.close();
        out.close();
    }
    else
    {
        assert(argc == 2);
        ifstream in(argv[0]);
        in.exceptions(ios::failbit);

        read_header(in, data, coders);
        decompressor d(in);
        decompress(d, data, coders);
        in.close();
        dequantise(data);
        writeout(argv[1], data);
    }
    return 0;
}
